@@ -19,6 +19,7 @@ extern "C" {
 #include <wchar.h>
 #include <stdio.h>
 #include <stdarg.h>
+#include <errno.h>
 
 #ifdef _MSC_VER
 typedef signed char int8_t;
@@ -30,6 +31,11 @@ typedef __int64 int64_t;
 #endif
 
 #ifdef _WIN32
+#ifndef UNICODE
+#define UNICODE
+#endif
+#define WIN32_LEAN_AND_MEAN
+#include <windows.h>
 #define EXPORT __declspec(dllexport)
 #else
 #define EXPORT
@@ -66,6 +72,15 @@ struct CheckFieldAlignment {
 
 static int _callCount;
 
+EXPORT void
+setLastError(int err) {
+#ifdef _WIN32  
+  SetLastError(err);
+#else
+  errno = err;
+#endif
+}
+
 EXPORT int
 callCount() {
   return ++_callCount;
@@ -331,6 +346,11 @@ incrementInt32ByReference(int32_t *arg) {
   if (arg) ++*arg;
 }
 
+EXPORT void 
+incrementNativeLongByReference(long *arg) {
+  if (arg) ++*arg;
+}
+
 EXPORT void 
 incrementInt64ByReference(int64_t *arg) {
   if (arg) ++*arg;
@@ -553,6 +573,28 @@ callInt32Callback(int32_t (*func)(int32_t arg, int32_t arg2),
   return (*func)(arg, arg2);
 }
 
+EXPORT int32_t 
+callInt32CallbackRepeatedly(int32_t (*func)(int32_t arg, int32_t arg2),
+                            int32_t arg, int32_t arg2, int32_t count) {
+  int i;
+  int sum = 0;
+  for (i=0;i < count;i++) {
+    sum += (*func)(arg, arg2);
+  }
+  return sum;
+}
+
+EXPORT long 
+callLongCallbackRepeatedly(long (*func)(long arg, long arg2),
+                           long arg, long arg2, int32_t count) {
+  int i;
+  long sum = 0;
+  for (i=0;i < count;i++) {
+    sum += (*func)(arg, arg2);
+  }
+  return sum;
+}
+
 EXPORT long 
 callNativeLongCallback(long (*func)(long arg, long arg2),
                        long arg, long arg2) {
@@ -670,6 +712,24 @@ fillInt64Buffer(int64_t *buf, int len, int64_t value) {
   return len;
 }
 
+EXPORT int32_t
+fillFloatBuffer(float *buf, int len, float value) {
+  int i;
+  for (i=0;i < len;i++) {
+    buf[i] = value;
+  }
+  return len;
+}
+
+EXPORT int32_t
+fillDoubleBuffer(double *buf, int len, double value) {
+  int i;
+  for (i=0;i < len;i++) {
+    buf[i] = value;
+  }
+  return len;
+}
+
 EXPORT int32_t
 addInt32VarArgs(const char *fmt, ...) {
   va_list ap;
@@ -768,7 +828,7 @@ callInt32StdCallCallback(int32_t (__stdcall *func)(int32_t arg, int32_t arg2),
 #include <jni.h>
 #include <math.h>
 JNIEXPORT jdouble JNICALL
-Java_com_sun_jna_RawTest_00024JNI_cos(JNIEnv *env, jclass cls, jdouble x) {
+Java_com_sun_jna_DirectTest_00024JNI_cos(JNIEnv *env, jclass cls, jdouble x) {
   return cos(x);
 }
 
