@@ -19,10 +19,22 @@ import java.util.Map;
  * An abstraction for a native function pointer.  An instance of 
  * <code>Function</code> represents a pointer to some native function.  
  * {@link #invoke(Class,Object[],Map)} is the primary means to call
- * the function. 
+ * the function. <p/>
+ * <a name=callflags></a>
+ * Function call behavior may be modified by passing one of the following call
+ * flags: 
+ * <ul>
+ * <li>{@link Function#C_CONVENTION} Use C calling convention (default)
+ * <li>{@link Function#ALT_CONVENTION} Use alternate calling convention (e.g. stdcall)
+ * <li>{@link Function#THROW_LAST_ERROR} Throw a {@link LastErrorException} if
+ * the native function sets the system error to a non-zero value (errno or
+ * GetLastError).  Setting this flag will cause the system error to be cleared
+ * prior to native function invocation.
+ * </ul>
  *
  * @author Sheng Liang, originator
  * @author Todd Fast, suitability modifications
+ * @author Timothy Wall
  * @see Pointer
  */
 public class Function extends Pointer {
@@ -45,6 +57,10 @@ public class Function extends Pointer {
     /** First alternate convention (currently used only for w32 stdcall). */
     public static final int ALT_CONVENTION = 1;
 
+    private static final int MASK_CC = 0x3;
+    /** Whether to throw an exception if last error is non-zero after call. */
+    public static final int THROW_LAST_ERROR = (1<<2);
+
     static final Integer INTEGER_TRUE = new Integer(-1);
     static final Integer INTEGER_FALSE = new Integer(0);
 
@@ -69,30 +85,64 @@ public class Function extends Pointer {
     
     /**
      * Obtain a <code>Function</code> representing a native 
-     * function that follows a given calling convention.
+     * function.
      * 
      * <p>The allocated instance represents a pointer to the named native 
-     * function from the named library, called with the named calling 
-     * convention.
+     * function from the named library.
      *
      * @param   libraryName
      *                  Library in which to find the function
      * @param   functionName
      *                  Name of the native function to be linked with
-     * @param   callConvention
-     *                  Call convention used by the native function
+     * @param   callFlags
+     *                  Function <a href="#callflags">call flags</a>
+     *                  
      * @throws {@link UnsatisfiedLinkError} if the library is not found or
      * the given function name is not found within the library.
      */
-    public static Function getFunction(String libraryName, String functionName, int callConvention) {
-        return NativeLibrary.getInstance(libraryName).getFunction(functionName, callConvention);
+    public static Function getFunction(String libraryName, String functionName, int callFlags) {
+        return NativeLibrary.getInstance(libraryName).getFunction(functionName, callFlags);
     }
     
-    // Keep a reference to the NativeLibrary so it does not get garbage collected
-    // until the function is
+    /**
+     * Obtain a <code>Function</code> representing a native 
+     * function pointer.  In general, this function should be used by dynamic
+     * languages; Java code should allow JNA to bind to a specific Callback
+     * interface instead by defining a return type or Structure field type.
+     * 
+     * <p>The allocated instance represents a pointer to the native 
+     * function pointer.
+     *
+     * @param   p       Native function pointer
+     */
+    public static Function getFunction(Pointer p) {
+        return getFunction(p, 0);
+    }
+
+    /**
+     * Obtain a <code>Function</code> representing a native 
+     * function pointer.  In general, this function should be used by dynamic
+     * languages; Java code should allow JNA to bind to a specific Callback
+     * interface instead by defining a return type or Structure field type. 
+     * 
+     * <p>The allocated instance represents a pointer to the native 
+     * function pointer.
+     *
+     * @param   p
+     *                  Native function pointer
+     * @param   callFlags
+     *                  Function <a href="#callflags">call flags</a>
+     */
+    public static Function getFunction(Pointer p, int callFlags) {
+        return new Function(p, callFlags);
+    }
+
+    // Keep a reference to the NativeLibrary so it does not get garbage
+    // collected until the function is
     private NativeLibrary library;
     private final String functionName;
-    final int callingConvention;
+    int callFlags;
+    final Map options;
 
     /** For internal JNA use. */
     static final String OPTION_INVOKING_METHOD = "invoking-method";
@@ -109,18 +159,19 @@ public class Function extends Pointer {
      *                 {@link NativeLibrary} in which to find the function
      * @param  functionName
      *                 Name of the native function to be linked with
-     * @param  callingConvention
-     *                 Calling convention used by the native function
+     * @param  callFlags
+     *                 Function <a href="#callflags">call flags</a>
      * @throws {@link UnsatisfiedLinkError} if the given function name is
      * not found within the library.
      */
-    Function(NativeLibrary library, String functionName, int callingConvention) {
-        checkCallingConvention(callingConvention);
+    Function(NativeLibrary library, String functionName, int callFlags) {
+        checkCallingConvention(callFlags & MASK_CC);
         if (functionName == null)
             throw new NullPointerException("Function name must not be null");
         this.library = library;
         this.functionName = functionName;
-        this.callingConvention = callingConvention;
+        this.callFlags = callFlags;
+        this.options = library.options;
         try {
             this.peer = library.getSymbolAddress(functionName);
         }
@@ -141,18 +192,19 @@ public class Function extends Pointer {
      *
      * @param  functionAddress
      *                 Address of the native function 
-     * @param  callingConvention
-     *                 Calling convention used by the native function
+     * @param  callFlags
+     *                 Function <a href="#callflags">call flags</a>
      */
-    Function(Pointer functionAddress, int callingConvention) {
-        checkCallingConvention(callingConvention);
+    Function(Pointer functionAddress, int callFlags) {
+        checkCallingConvention(callFlags & MASK_CC);
         if (functionAddress == null
             || functionAddress.peer == 0) {
             throw new NullPointerException("Function address may not be null");
         }
         this.functionName = functionAddress.toString();
-        this.callingConvention = callingConvention;
+        this.callFlags = callFlags;
         this.peer = functionAddress.peer;
+        this.options = Collections.EMPTY_MAP;
     }
     
     private void checkCallingConvention(int convention)
@@ -173,14 +225,14 @@ public class Function extends Pointer {
 
 
     public int getCallingConvention() {
-        return callingConvention;
+        return callFlags & MASK_CC;
     }
 
     /** Invoke the native function with the given arguments, returning the
      * native result as an Object.
      */
     public Object invoke(Class returnType, Object[] inArgs) {
-        return invoke(returnType, inArgs, Collections.EMPTY_MAP);
+        return invoke(returnType, inArgs, this.options);
     }    
     
     /** Invoke the native function with the given arguments, returning the
@@ -243,9 +295,7 @@ public class Function extends Pointer {
                     continue;
                 if (inArg instanceof Structure) {
                     if (!(inArg instanceof Structure.ByValue)) {
-                        if (((Structure)inArg).getAutoRead()) {
-                            ((Structure)inArg).read();
-                        }
+                        ((Structure)inArg).autoRead();
                     }
                 }
                 else if (args[i] instanceof PostCallRead) {
@@ -263,12 +313,7 @@ public class Function extends Pointer {
                     }
                 }
                 else if (Structure[].class.isAssignableFrom(inArg.getClass())) {
-                    Structure[] ss = (Structure[])inArg;
-                    for (int si=0;si < ss.length;si++) {
-                        if (ss[si].getAutoRead()) {
-                            ss[si].read();
-                        }
-                    }
+                    Structure.autoRead((Structure[])inArg);
                 }
             }
         }
@@ -280,81 +325,77 @@ public class Function extends Pointer {
     Object invoke(Object[] args, Class returnType, boolean allowObjects) {
         Object result = null;
         if (returnType == null || returnType==void.class || returnType==Void.class) {
-            invokeVoid(callingConvention, args);
+            invokeVoid(callFlags, args);
             result = null;
         }
         else if (returnType==boolean.class || returnType==Boolean.class) {
-            result = valueOf(invokeInt(callingConvention, args) != 0);
+            result = valueOf(invokeInt(callFlags, args) != 0);
         }
         else if (returnType==byte.class || returnType==Byte.class) {
-            result = new Byte((byte)invokeInt(callingConvention, args));
+            result = new Byte((byte)invokeInt(callFlags, args));
         }
         else if (returnType==short.class || returnType==Short.class) {
-            result = new Short((short)invokeInt(callingConvention, args));
+            result = new Short((short)invokeInt(callFlags, args));
         }
         else if (returnType==char.class || returnType==Character.class) {
-            result = new Character((char)invokeInt(callingConvention, args));
+            result = new Character((char)invokeInt(callFlags, args));
         }
         else if (returnType==int.class || returnType==Integer.class) {
-            result = new Integer(invokeInt(callingConvention, args));
+            result = new Integer(invokeInt(callFlags, args));
         }
         else if (returnType==long.class || returnType==Long.class) {
-            result = new Long(invokeLong(callingConvention, args));
+            result = new Long(invokeLong(callFlags, args));
         }
         else if (returnType==float.class || returnType==Float.class) {
-            result = new Float(invokeFloat(callingConvention, args));
+            result = new Float(invokeFloat(callFlags, args));
         }
         else if (returnType==double.class || returnType==Double.class) {
-            result = new Double(invokeDouble(callingConvention, args));
+            result = new Double(invokeDouble(callFlags, args));
         }
         else if (returnType==String.class) {
-            result = invokeString(callingConvention, args, false);
+            result = invokeString(callFlags, args, false);
         }
         else if (returnType==WString.class) {
-            String s = invokeString(callingConvention, args, true);
+            String s = invokeString(callFlags, args, true);
             if (s != null) {
                 result = new WString(s);
             }
         }
         else if (Pointer.class.isAssignableFrom(returnType)) {
-            result = invokePointer(callingConvention, args);
+            result = invokePointer(callFlags, args);
         }
         else if (Structure.class.isAssignableFrom(returnType)) {
             if (Structure.ByValue.class.isAssignableFrom(returnType)) {
                 Structure s = 
-                    invokeStructure(callingConvention, args, 
+                    invokeStructure(callFlags, args, 
                                     Structure.newInstance(returnType));
-                if (s.getAutoRead()) {
-                    s.read();
-                }
+                s.autoRead();
                 result = s;
             }
             else {
-                result = invokePointer(callingConvention, args);
+                result = invokePointer(callFlags, args);
                 if (result != null) {
                     Structure s = Structure.newInstance(returnType);
                     s.useMemory((Pointer)result);
-                    if (s.getAutoRead()) {
-                        s.read();
-                    }
+                    s.autoRead();
                     result = s;
                 }
             }
         }
         else if (Callback.class.isAssignableFrom(returnType)) {
-            result = invokePointer(callingConvention, args);
+            result = invokePointer(callFlags, args);
             if (result != null) {
                 result = CallbackReference.getCallback(returnType, (Pointer)result);
             }
         }
         else if (returnType==String[].class) {
-            Pointer p = invokePointer(callingConvention, args);
+            Pointer p = invokePointer(callFlags, args);
             if (p != null) {
                 result = p.getStringArray(0);
             }
         }
         else if (returnType==WString[].class) {
-            Pointer p = invokePointer(callingConvention, args);
+            Pointer p = invokePointer(callFlags, args);
             if (p != null) {
                 String[] arr = p.getStringArray(0, true);
                 WString[] warr = new WString[arr.length];
@@ -365,18 +406,18 @@ public class Function extends Pointer {
             }
         }
         else if (returnType==Pointer[].class) {
-            Pointer p = invokePointer(callingConvention, args);
+            Pointer p = invokePointer(callFlags, args);
             if (p != null) {
                 result = p.getPointerArray(0);
             }
         }
         else if (allowObjects) {
-            result = invokeObject(callingConvention, args);
+            result = invokeObject(callFlags, args);
             if (result != null
                 && !returnType.isAssignableFrom(result.getClass())) {
-                throw new IllegalArgumentException("Return type " + returnType
-                                                   + " does not match result "
-                                                   + result.getClass());
+                throw new ClassCastException("Return type " + returnType
+                                             + " does not match result "
+                                             + result.getClass());
             }
         }
         else {
@@ -418,9 +459,7 @@ public class Function extends Pointer {
         // Convert Structures to native pointers 
         if (arg instanceof Structure) {
             Structure struct = (Structure)arg;
-            if (struct.getAutoWrite()) {
-                struct.write();
-            }
+            struct.autoWrite();
             if (struct instanceof Structure.ByValue) {
             	// Double-check against the method signature, if available
                 Class ptype = struct.getClass();
@@ -476,6 +515,9 @@ public class Function extends Pointer {
         else if (Pointer[].class == argClass) {
             return new PointerArray((Pointer[])arg);
         }
+        else if (NativeMapped[].class.isAssignableFrom(argClass)) {
+            return new NativeMappedArray((NativeMapped[])arg);
+        }
         else if (Structure[].class.isAssignableFrom(argClass)) {
             Structure[] ss = (Structure[])arg;
             Class type = argClass.getComponentType();
@@ -483,7 +525,7 @@ public class Function extends Pointer {
             if (byRef) {
                 Pointer[] pointers = new Pointer[ss.length + 1];
                 for (int i=0;i < ss.length;i++) {
-                	pointers[i] = ss[i] != null ? ss[i].getPointer() : null;
+                    pointers[i] = ss[i] != null ? ss[i].getPointer() : null;
                 }
                 return new PointerArray(pointers);
             }
@@ -491,39 +533,15 @@ public class Function extends Pointer {
                 throw new IllegalArgumentException("Structure array must have non-zero length");
             }
             else if (ss[0] == null) {
-                // Initialize uninitialized arrays of Structure to point
-                // to a single block of memory
-                Structure struct = Structure.newInstance(type);
-                int size = struct.size();
-                Memory m = new Memory(size * ss.length);
-                struct.useMemory(m);
-                Structure[] tmp = struct.toArray(ss.length);
-                for (int si=0;si < ss.length;si++) {
-                    ss[si] = tmp[si];
-                }
+                Structure.newInstance(type).toArray(ss);
                 return ss[0].getPointer();
             }
             else {
-                Pointer base = ss[0].getPointer();
-                int size = ss[0].size();
-                if (ss[0].getAutoWrite()) {
-                    ss[0].write();
-                }
-                for (int si=1;si < ss.length;si++) {
-                    if (ss[si].getPointer().peer != base.peer + size*si) {
-                        String msg = "Structure array elements must use"
-                            + " contiguous memory (at element index " + si + ")";     
-                        throw new IllegalArgumentException(msg);
-                    }
-                    if (ss[si].getAutoWrite()) {
-                        ss[si].write();
-                    }
-                }
-                return base;
+                Structure.autoWrite(ss);
+                return ss[0].getPointer();
             }
         }
         else if (argClass.isArray()){
-            // TODO: handle array of NativeMapped
             throw new IllegalArgumentException("Unsupported array argument type: " 
                                                + argClass.getComponentType());
         }
@@ -547,22 +565,22 @@ public class Function extends Pointer {
     /**
      * Call the native function being represented by this object
      *
-     * @param   callingConvention calling convention to be used
+     * @param   callFlags calling convention to be used
      * @param	args
      *			Arguments to pass to the native function
      * @return	The value returned by the target native function
      */
-    private  native int invokeInt(int callingConvention, Object[] args);
+    private  native int invokeInt(int callFlags, Object[] args);
 
     /**
      * Call the native function being represented by this object
      *
-     * @param   callingConvention calling convention to be used
+     * @param   callFlags calling convention to be used
      * @param	args
      *			Arguments to pass to the native function
      * @return	The value returned by the target native function
      */
-    private native long invokeLong(int callingConvention, Object[] args);
+    private native long invokeLong(int callFlags, Object[] args);
 
     /**
      * Call the native function being represented by this object
@@ -578,44 +596,44 @@ public class Function extends Pointer {
     /**
      * Call the native function being represented by this object
      *
-     * @param   callingConvention calling convention to be used
+     * @param   callFlags calling convention to be used
      * @param	args
      *			Arguments to pass to the native function
      */
-    private native void invokeVoid(int callingConvention, Object[] args);
+    private native void invokeVoid(int callFlags, Object[] args);
 
     /**
      * Call the native function being represented by this object
      *
-     * @param   callingConvention calling convention to be used
+     * @param   callFlags calling convention to be used
      * @param	args
      *			Arguments to pass to the native function
      * @return	The value returned by the target native function
      */
-    private native float invokeFloat(int callingConvention, Object[] args);
+    private native float invokeFloat(int callFlags, Object[] args);
 
     /**
      * Call the native function being represented by this object
      *
-     * @param   callingConvention calling convention to be used
+     * @param   callFlags calling convention to be used
      * @param	args
      *			Arguments to pass to the native function
      * @return	The value returned by the target native function
      */
-    private native double invokeDouble(int callingConvention, Object[] args);
+    private native double invokeDouble(int callFlags, Object[] args);
 
     /**
      * Call the native function being represented by this object
      *
-     * @param   callingConvention calling convention to be used
+     * @param   callFlags calling convention to be used
      * @param	args
      *			Arguments to pass to the native function
      * @param   wide whether the native string uses <code>wchar_t</code>;
      * if false, <code>char</code> is assumed
      * @return	The value returned by the target native function, as a String
      */
-    private String invokeString(int callingConvention, Object[] args, boolean wide) {
-        Pointer ptr = invokePointer(callingConvention, args);
+    private String invokeString(int callFlags, Object[] args, boolean wide) {
+        Pointer ptr = invokePointer(callFlags, args);
         String s = null;
         if (ptr != null) {
             if (wide)
@@ -629,36 +647,36 @@ public class Function extends Pointer {
     /**
      * Call the native function being represented by this object
      *
-     * @param   callingConvention calling convention to be used
+     * @param   callFlags calling convention to be used
      * @param	args
      *			Arguments to pass to the native function
      * @return	The native pointer returned by the target native function
      */
-    private native Pointer invokePointer(int callingConvention, Object[] args);
+    private native Pointer invokePointer(int callFlags, Object[] args);
     
     /**
      * Call the native function being represented by this object, returning
      * a struct by value.
      *
-     * @param   callingConvention calling convention to be used
+     * @param   callFlags calling convention to be used
      * @param   args
      *          Arguments to pass to the native function
      * @param   result Pre-allocated structure to hold the result
      * @return  The passed-in struct argument
      */
-    private native Structure invokeStructure(int callingConvention, Object[] args,
+    private native Structure invokeStructure(int callFlags, Object[] args,
                                              Structure result);
 
     /**
      * Call the native function being represented by this object, returning
      * a Java <code>Object</code>.
      *
-     * @param   callingConvention calling convention to be used
+     * @param   callFlags calling convention to be used
      * @param   args
      *          Arguments to pass to the native function
      * @return  The returned Java <code>Object</code>
      */
-    private native Object invokeObject(int callingConvention, Object[] args);
+    private native Object invokeObject(int callFlags, Object[] args);
 
     /** Provide a human-readable representation of this object. */
     public String toString() {
@@ -730,9 +748,12 @@ public class Function extends Pointer {
      * and calling convention.
      */
     public boolean equals(Object o) {
-        if (o instanceof Function) {
+        if (o == this) return true;
+        if (o == null) return false;
+        if (o.getClass() == getClass()) {
             Function other = (Function)o;
-            return other.callingConvention == this.callingConvention
+            return other.callFlags == this.callFlags
+                && other.options.equals(this.options)
                 && other.peer == this.peer;
         }
         return false;
@@ -784,6 +805,19 @@ public class Function extends Pointer {
         return false;
     }
     
+    private static class NativeMappedArray extends Memory implements PostCallRead {
+        private final NativeMapped[] original;
+        public NativeMappedArray(NativeMapped[] arg) {
+            super(Native.getNativeSize(arg.getClass(), arg));
+            this.original = arg;
+            Class nativeType = arg.getClass().getComponentType();
+            setValue(0, original, original.getClass());
+        }
+        public void read() {
+            getValue(0, original.getClass(), original);
+        }
+    }
+
     private static class PointerArray extends Memory implements PostCallRead {
         private final Pointer[] original;
         public PointerArray(Pointer[] arg) {
@@ -795,9 +829,7 @@ public class Function extends Pointer {
             setPointer(Pointer.SIZE*arg.length, null);
         }
         public void read() {
-            for (int i=0;i < original.length;i++) {
-                original[i] = getPointer(i * Pointer.SIZE);
-            }
+            read(0, original, 0, original.length);
         }
     }
     
