@@ -11,7 +11,8 @@
 package com.sun.jna;
 
 import java.nio.ByteBuffer;
-
+import java.util.Map;
+import java.util.WeakHashMap;
 
 /**
  * A <code>Pointer</code> to memory obtained from the native heap via a 
@@ -36,8 +37,17 @@ import java.nio.ByteBuffer;
  */
 public class Memory extends Pointer {
 
+    private static Map buffers = new WeakHashMap();
+
     protected long size; // Size of the malloc'ed space
 
+    /** Force cleanup of memory that has associated NIO Buffers which have
+        been GC'd.
+    */
+    public static void purge() {
+        buffers.size();
+    }
+
     /** Provide a view into the original memory. */
     private class SharedMemory extends Memory {
         public SharedMemory(long offset) {
@@ -105,17 +115,22 @@ public class Memory extends Pointer {
         if (byteBoundary <= 0) {
             throw new IllegalArgumentException("Byte boundary must be positive: " + byteBoundary);
         }
-        long mask = ~((long)byteBoundary - 1);
-
-        if ((peer & ~mask) != peer) {
-            long newPeer = (peer + ~mask) & mask;
-            long newSize = peer + size - newPeer;
-            if (newSize <= 0) {
-                throw new IllegalArgumentException("Insufficient memory to align to the requested boundary");
+        for (int i=0;i < 32;i++) {
+            if (byteBoundary == (1<<i)) {
+                long mask = ~((long)byteBoundary - 1);
+                
+                if ((peer & mask) != peer) {
+                    long newPeer = (peer + byteBoundary - 1) & mask;
+                    long newSize = peer + size - newPeer;
+                    if (newSize <= 0) {
+                        throw new IllegalArgumentException("Insufficient memory to align to the requested boundary");
+                    }
+                    return (Memory)share(newPeer - peer, newSize);
+                }
+                return this;
             }
-            return (Memory)share(newPeer - peer, newSize);
         }
-        return this;
+        throw new IllegalArgumentException("Byte boundary must be a power of two");
     }
 
     protected void finalize() {
@@ -128,15 +143,25 @@ public class Memory extends Pointer {
         clear(size);
     }
 
-    /** Returns false if the memory has been freed. */
+    /** @deprecated use {@link #valid()} instead. */
     public boolean isValid() {
+        return valid();
+    }
+
+    /** Returns false if the memory has been freed. */
+    public boolean valid() {
         return peer != 0;
     }
 
-    public long getSize() {
+    public long size() {
         return size;
     }
 
+    /** @deprecated use {@link #size()} instead. */
+    public long getSize() {
+        return size();
+    }
+
 
     /**
      * Check that indirection won't cause us to write outside the 
@@ -487,7 +512,11 @@ public class Memory extends Pointer {
      */
     public ByteBuffer getByteBuffer(long offset, long length) {
         boundsCheck(offset, length);
-        return super.getByteBuffer(offset, length);
+        ByteBuffer b = super.getByteBuffer(offset, length);
+        // Ensure this Memory object will not be GC'd (and its memory freed)
+        // if the Buffer is still extant.
+        buffers.put(b, this);
+        return b;
     }
 
     /**
