@@ -17,6 +17,7 @@ package com.sun.jna;
 import java.io.File;
 import java.io.FilenameFilter;
 import java.lang.ref.WeakReference;
+import java.lang.reflect.Method;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collections;
@@ -29,7 +30,8 @@ import java.util.StringTokenizer;
 
 /**
  * Provides management of native library resources.  One instance of this
- * class corresponds to a single loaded native library.
+ * class corresponds to a single loaded native library.  May also be used
+ * to map to the current process (see {@link NativeLibrary#getProcess()}).
  * <p>
  * <b>Library Search Paths</b>
  * A search for a given library will scan the following locations:
@@ -45,14 +47,12 @@ public class NativeLibrary {
     private final String libraryName;
     private final String libraryPath;
     private final Map functions = new HashMap();
-    private final int callingConvention;
-    private final Map options;
+    final int callFlags;
+    final Map options;
 
-    private static WeakReference currentProcess;
     private static final Map libraries = new HashMap();
     private static final Map searchPaths = Collections.synchronizedMap(new HashMap());
     private static final List librarySearchPath = new LinkedList();
-    private static final List userSearchPath = new LinkedList();
 
     static {
         // Force initialization of native library
@@ -60,6 +60,10 @@ public class NativeLibrary {
             throw new Error("Native library not initialized");
     }
 
+    private static String functionKey(String name, int flags) {
+        return name + "|" + flags;
+    }
+
     private NativeLibrary(String libraryName, String libraryPath, long handle, Map options) {
         this.libraryName = getLibraryName(libraryName);
         this.libraryPath = libraryPath;
@@ -67,7 +71,7 @@ public class NativeLibrary {
         Object option = options.get(Library.OPTION_CALLING_CONVENTION);
         int callingConvention = option instanceof Integer
             ? ((Integer)option).intValue() : Function.C_CONVENTION;
-        this.callingConvention = callingConvention;
+        this.callFlags = callingConvention;
         this.options = options;
 
         // Special workaround for w32 kernel32.GetLastError
@@ -79,7 +83,7 @@ public class NativeLibrary {
                         return new Integer(Native.getLastError());
                     }
                 };
-                functions.put("GetLastError", f);
+                functions.put(functionKey("GetLastError", callFlags), f);
             }
         }
     }
@@ -104,13 +108,13 @@ public class NativeLibrary {
             }
         }
         
-        searchPath.addAll(userSearchPath);
+        searchPath.addAll(initPaths("jna.library.path"));
         String libraryPath = findLibraryPath(libraryName, searchPath);
         long handle = 0;
         //
         // Only search user specified paths first.  This will also fall back
-        // to dlopen/LoadLibrary() since findLibraryPath returns the mapped name
-        // if it cannot find the library.
+        // to dlopen/LoadLibrary() since findLibraryPath returns the mapped
+        // name if it cannot find the library.
         //
         try {
             handle = open(libraryPath);
@@ -132,7 +136,9 @@ public class NativeLibrary {
                 //
                 libraryPath = matchLibrary(libraryName, searchPath);
                 if (libraryPath != null) {
-                    try { handle = open(libraryPath); }
+                    try {
+                        handle = open(libraryPath);
+                    }
                     catch(UnsatisfiedLinkError e2) { e = e2; }
                 }
             }
@@ -141,7 +147,9 @@ public class NativeLibrary {
                 libraryPath = "/System/Library/Frameworks/" + libraryName
                     + ".framework/" + libraryName;
                 if (new File(libraryPath).exists()) {
-                    try { handle = open(libraryPath); }
+                    try { 
+                        handle = open(libraryPath);
+                    }
                     catch(UnsatisfiedLinkError e2) { e = e2; }
                 }
             }
@@ -206,46 +214,62 @@ public class NativeLibrary {
      * Library}).
      */
     public static final NativeLibrary getInstance(String libraryName, Map options) {
-        if (libraryName == null)
-            throw new NullPointerException("Library name may not be null");
+        options = new HashMap(options);
+        if (options.get(Library.OPTION_CALLING_CONVENTION) == null) {
+            options.put(Library.OPTION_CALLING_CONVENTION, new Integer(Function.C_CONVENTION));
+        }
 
+        // Use current process to load libraries we know are already
+        // loaded by the VM to ensure we get the correct version
+        if (Platform.isLinux() && "c".equals(libraryName)) {
+            libraryName = null;
+        }
         synchronized (libraries) {
-            WeakReference ref = (WeakReference)libraries.get(libraryName);
+            WeakReference ref = (WeakReference)libraries.get(libraryName + options);
             NativeLibrary library = ref != null ? (NativeLibrary)ref.get() : null;
+
             if (library == null) {
-                library = loadLibrary(libraryName, options);
+                if (libraryName == null) {
+                    library = new NativeLibrary("<process>", null, open(null), options);
+                }
+                else {
+                    library = loadLibrary(libraryName, options);
+                }
                 ref = new WeakReference(library);
-                libraries.put(library.getName(), ref);
-                libraries.put(library.getFile().getAbsolutePath(), ref);
-                libraries.put(library.getFile().getName(), ref);
+                libraries.put(library.getName() + options, ref);
+                File file = library.getFile();
+                if (file != null) {
+                    libraries.put(file.getAbsolutePath() + options, ref);
+                    libraries.put(file.getName() + options, ref);
+                }
             }
             return library;
         }
     }
 
     /**
-     * Returns an instance of NativeLibrary which refers to the current process.
-     * This is useful for accessing functions which were already mapped by some
-     * other mechanism, without having to reference or even know the exact
-     * name of the native library.
+     * Returns an instance of NativeLibrary which refers to the current
+     * process.  This is useful for accessing functions which were already
+     * mapped by some other mechanism, without having to reference or even
+     * know the exact name of the native library.
      */
     public static synchronized final NativeLibrary getProcess() {
-        NativeLibrary library = null;
-        if (currentProcess != null) {
-            library = (NativeLibrary) currentProcess.get();
-        }
-
-        if (library == null) {
-           long handle = open(null);
-           library = new NativeLibrary("<process>", null, handle, Collections.EMPTY_MAP);
-           currentProcess = new WeakReference(library);
-        }
-
-        return library;
+        return getInstance(null);
     }
 
     /**
-     * Add a path to search for the specified library, ahead of any system paths
+     * Returns an instance of NativeLibrary which refers to the current
+     * process.  This is useful for accessing functions which were already
+     * mapped by some other mechanism, without having to reference or even
+     * know the exact name of the native library.
+     */
+    public static synchronized final NativeLibrary getProcess(Map options) {
+        return getInstance(null, options);
+    }
+
+    /**
+     * Add a path to search for the specified library, ahead of any system
+     * paths.
      *
      * @param libraryName The name of the library to use the path for
      * @param path The path to use when trying to load the library
@@ -263,41 +287,62 @@ public class NativeLibrary {
     }
     /**
      * Create a new {@link Function} that is linked with a native
-     * function that follows the standard "C" calling convention.
+     * function that follows the NativeLibrary's calling convention.
      *
      * <p>The allocated instance represents a pointer to the named native
-     * function from the library, called with the standard "C" calling
-     * convention.
+     * function from the library.
      *
      * @param	functionName
      *			Name of the native function to be linked with
      * @throws   UnsatisfiedLinkError if the function is not found
      */
     public Function getFunction(String functionName) {
-        return getFunction(functionName, callingConvention);
+        return getFunction(functionName, callFlags);
+    }
+
+    /**
+     * Create a new {@link Function} that is linked with a native
+     * function that follows the NativeLibrary's calling convention.
+     *
+     * <p>The allocated instance represents a pointer to the named native
+     * function from the library.
+     *
+     * @param	name
+     *			Name of the native function to be linked with
+     * @param	method
+     *			Method to which the native function is to be mapped
+     * @throws   UnsatisfiedLinkError if the function is not found
+     */
+    Function getFunction(String name, Method method) {
+        int flags = this.callFlags;
+        Class[] etypes = method.getExceptionTypes();
+        for (int i=0;i < etypes.length;i++) {
+            if (LastErrorException.class.isAssignableFrom(etypes[i])) {
+                flags |= Function.THROW_LAST_ERROR;
+            }
+        }
+        return getFunction(name, flags);
     }
 
     /**
      * Create a new  @{link Function} that is linked with a native
-     * function that follows a given calling convention.
-     *
-     * <p>The allocated instance represents a pointer to the named native
-     * function from the library, called with the named calling convention.
+     * function that follows a given calling flags.
      *
      * @param	functionName
      *			Name of the native function to be linked with
-     * @param	callingConvention
-     *			Calling convention used by the native function
+     * @param	callFlags
+     *			Flags affecting the function invocation
      * @throws   UnsatisfiedLinkError if the function is not found
      */
-    public Function getFunction(String functionName, int callingConvention) {
+    public Function getFunction(String functionName, int callFlags) {
         if (functionName == null)
             throw new NullPointerException("Function name may not be null");
         synchronized (functions) {
-            Function function = (Function) functions.get(functionName);
+            String key = functionKey(functionName, callFlags);
+            Function function = (Function) functions.get(key);
             if (function == null) {
-                function = new Function(this, functionName, callingConvention);
-                functions.put(functionName, function);
+                function = new Function(this, functionName, callFlags);
+                functions.put(key, function);
             }
             return function;
         }
@@ -357,11 +402,11 @@ public class NativeLibrary {
 
     public void dispose() {
         synchronized(libraries) {
-            libraries.remove(getName());
-            File path = getFile();
-            if (path != null) {
-                libraries.remove(path.getAbsolutePath());
-                libraries.remove(path.getName());
+            libraries.remove(getName() + options);
+            File file = getFile();
+            if (file != null) {
+                libraries.remove(file.getAbsolutePath() + options);
+                libraries.remove(file.getName() + options);
             }
         }
         synchronized(this) {
@@ -405,10 +450,21 @@ public class NativeLibrary {
 
         // Search in the JNA paths for it
         for (Iterator it = searchPath.iterator(); it.hasNext(); ) {
-            File file = new File(new File((String) it.next()), name);
+            String path = (String)it.next();
+            File file = new File(path, name);
             if (file.exists()) {
                 return file.getAbsolutePath();
             }
+            if (Platform.isMac()) {
+                // Native libraries delivered via JNLP class loader
+                // may require a .jnilib extension to be found
+                if (name.endsWith(".dylib")) {
+                    file = new File(path, name.substring(0, name.lastIndexOf(".dylib")) + ".jnilib");
+                    if (file.exists()) {
+                        return file.getAbsolutePath();
+                    }
+                }
+            }
         }
 
         //
@@ -440,6 +496,11 @@ public class NativeLibrary {
                 return libName;
             }
         }
+        else if (Platform.isWindows()) {
+        	if (libName.endsWith(".drv") || libName.endsWith(".dll")) {
+        		return libName;
+        	}
+        }
 
         return System.mapLibraryName(libName);
     }
@@ -535,7 +596,6 @@ public class NativeLibrary {
     private static native void close(long handle);
     private static native long findSymbol(long handle, String name);
     static {
-        userSearchPath.addAll(initPaths("jna.library.path"));
         String webstartPath = Native.getWebStartLibraryPath("jnidispatch");
         if (webstartPath != null) {
             librarySearchPath.add(webstartPath);
