@@ -1,4 +1,4 @@
-/* Copyright (c) 2007 Timothy Wall, All Rights Reserved
+/* Copyright (c) 2007-2009 Timothy Wall, All Rights Reserved
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -14,11 +14,13 @@ package com.sun.jna;
 
 import java.nio.Buffer;
 import java.nio.ByteBuffer;
+import java.nio.DoubleBuffer;
+import java.util.Arrays;
+import java.util.List;
 import java.util.Map;
 
 import junit.framework.TestCase;
 
-import com.sun.jna.StructureTest.VariableSizeTest.VariableSizedStructure;
 import com.sun.jna.ptr.ByteByReference;
 import com.sun.jna.ptr.IntByReference;
 import com.sun.jna.ptr.LongByReference;
@@ -26,6 +28,7 @@ import com.sun.jna.ptr.LongByReference;
 /** TODO: need more alignment tests, especially platform-specific behavior
  * @author twall@users.sf.net
  */
+@SuppressWarnings("unused")
 public class StructureTest extends TestCase {
 
     public static void main(java.lang.String[] argList) {
@@ -468,6 +471,35 @@ public class StructureTest extends TestCase {
         }
     }
 
+    public static class BadFieldStructure extends Structure {
+        public Object badField;
+    }
+    public void testUnsupportedField() {
+        class BadNestedStructure extends Structure {
+            public BadFieldStructure badStruct = new BadFieldStructure();
+        }
+        try {
+            new BadFieldStructure().size();
+            fail("Should throw IllegalArgumentException on bad field");
+        }
+        catch(IllegalArgumentException e) {
+            assertTrue("Exception should include field name",
+                       e.getMessage().indexOf("badField") != -1);
+        }
+        try {
+            new BadNestedStructure().size();
+            fail("Should throw IllegalArgumentException on bad field");
+        }
+        catch(IllegalArgumentException e) {
+            assertTrue("Exception should include enclosing type: " + e,
+                       e.getMessage().indexOf(BadNestedStructure.class.getName()) != -1);
+            assertTrue("Exception should include nested field name: " + e,
+                       e.getMessage().indexOf("badStruct") != -1);
+            assertTrue("Exception should include field name: " + e,
+                       e.getMessage().indexOf("badField") != -1);
+        }
+    }
+
     public void testToArray() {
         PublicTestStructure s = new PublicTestStructure();
         PublicTestStructure[] array = (PublicTestStructure[])s.toArray(1);
@@ -477,18 +509,36 @@ public class StructureTest extends TestCase {
         assertEquals("Structure memory should be expanded", 2, s.toArray(2).length);
     }
 
+    public void testByReferenceArraySync() {
+        PublicTestStructure.ByReference s = new PublicTestStructure.ByReference();
+        PublicTestStructure.ByReference[] array =
+            (PublicTestStructure.ByReference[])s.toArray(2);
+        class TestStructure extends Structure {
+            public PublicTestStructure.ByReference ref;
+        }
+        TestStructure ts = new TestStructure();
+        ts.ref = s;
+        final int VALUE = 42;
+        array[0].x = VALUE;
+        array[1].x = VALUE;
+        ts.write();
+
+        assertEquals("Array element not written: " + array[0],
+                     VALUE, array[0].getPointer().getInt(0));
+        assertEquals("Array element not written: " + array[1],
+                     VALUE, array[1].getPointer().getInt(0));
+
+        array[0].getPointer().setInt(4, VALUE);
+        array[1].getPointer().setInt(4, VALUE);
+        ts.read();
+
+        assertEquals("Array element not read: " + array[0], VALUE, array[0].y);
+        assertEquals("Array element not read: " + array[1], VALUE, array[1].y);
+    }
+
     static class CbStruct extends Structure {
         public Callback cb;
     }
-    static class CbStruct2 extends Structure {
-        public static interface TestCallback extends Callback {
-            int callback(int arg1, int arg2);
-        }
-        public TestCallback cb;
-    }
-    static interface CbTest extends Library {
-        public void setCallbackInStruct(CbStruct2 cbstruct);
-    }
     public void testCallbackWrite() {
         final CbStruct s = new CbStruct();
         s.cb = new Callback() {
@@ -504,21 +554,6 @@ public class StructureTest extends TestCase {
         assertEquals("Wrong trampoline", ref.getTrampoline(), func);
     }
 
-    public void testReadFunctionPointerAsCallback() {
-        CbStruct2 s = new CbStruct2();
-        CbTest lib = (CbTest)Native.loadLibrary("testlib", CbTest.class);
-        lib.setCallbackInStruct(s);
-        assertNotNull("Callback field not set", s.cb);
-    }
-
-    public void testCallProxiedFunctionPointer() {
-        CbStruct2 s = new CbStruct2();
-        CbTest lib = (CbTest)Native.loadLibrary("testlib", CbTest.class);
-        lib.setCallbackInStruct(s);
-        assertEquals("Proxy to native function pointer failed",
-                     3, s.cb.callback(1, 2));
-    }
-
     public void testUninitializedArrayField() {
         class UninitializedArrayFieldStructure extends Structure {
             public byte[] array;
@@ -564,6 +599,7 @@ public class StructureTest extends TestCase {
 
     class BufferStructure extends Structure {
         public Buffer buffer;
+        public DoubleBuffer dbuffer;
     }
     public void testBufferFieldWriteNULL() {
         BufferStructure bs = new BufferStructure();
@@ -572,13 +608,17 @@ public class StructureTest extends TestCase {
     public void testBufferFieldWriteNonNULL() {
         BufferStructure bs = new BufferStructure();
         bs.buffer = ByteBuffer.allocateDirect(16);
+        bs.dbuffer = ((ByteBuffer)bs.buffer).asDoubleBuffer();
         bs.write();
     }
     public void testBufferFieldReadUnchanged() {
         BufferStructure bs = new BufferStructure();
-        bs.buffer = ByteBuffer.allocateDirect(16);
+        Buffer b = ByteBuffer.allocateDirect(16);
+        bs.buffer = b;
+        bs.dbuffer = ((ByteBuffer)bs.buffer).asDoubleBuffer();
         bs.write();
         bs.read();
+        assertEquals("Buffer field should be unchanged", b, bs.buffer);
     }
     public void testBufferFieldReadChanged() {
         BufferStructure bs = new BufferStructure();
@@ -605,8 +645,10 @@ public class StructureTest extends TestCase {
     public void testBufferFieldReadChangedToNULL() {
         BufferStructure bs = new BufferStructure();
         bs.buffer = ByteBuffer.allocateDirect(16);
+        bs.dbuffer = ((ByteBuffer)bs.buffer).asDoubleBuffer();
         bs.read();
         assertNull("Structure Buffer field should be set null", bs.buffer);
+        assertNull("Structure DoubleBuffer field should be set null", bs.dbuffer);
     }
 
     public void testVolatileStructureField() {
@@ -693,7 +735,7 @@ public class StructureTest extends TestCase {
         s.s1 = new StructureTest.PublicTestStructure.ByReference();
         s.s1.x = -1;
         s.write();
-        assertEquals("Structure.ByReference not written automatically",
+        assertEquals("Structure.ByReference field not written automatically",
                      -1, s.s1.getPointer().getInt(0));
     }
 
@@ -724,6 +766,26 @@ public class StructureTest extends TestCase {
         assertEquals("New ref should be equivalent", ref, s.array[0]);
     }
 
+    public void testAutoReadWriteStructureByReferenceArrayField() {
+        class TestStructure extends Structure {
+            public PublicTestStructure.ByReference field;
+        }
+        TestStructure s = new TestStructure();
+        s.field = new PublicTestStructure.ByReference();
+        PublicTestStructure.ByReference[] array = 
+            (PublicTestStructure.ByReference[])s.field.toArray(2);
+        final int VALUE = -1;
+        array[1].x = VALUE;
+        s.write();
+        assertEquals("ByReference array member not auto-written",
+                     VALUE, array[1].getPointer().getInt(0));
+
+        array[1].getPointer().setInt(0, VALUE*2);
+        s.read();
+        assertEquals("ByReference array member not auto-read",
+                     VALUE*2, array[1].x);
+    }
+
     static class NestedTypeInfoStructure extends Structure {
         public static class Inner extends Structure {
             public int dummy;
@@ -801,46 +863,70 @@ public class StructureTest extends TestCase {
                    actual.matches(EXPECTED));
     }
 
-    public interface VariableSizeTest extends Library {
-        public static class VariableSizedStructure extends Structure {
-            public int length;
-            public byte[] buffer;
-            public VariableSizedStructure(String arg) {
-                length = arg.length() + 1;
-                buffer = new byte[length];
-                System.arraycopy(arg.getBytes(), 0, buffer, 0, arg.length());
-            }
-        }
-        String returnStringFromVariableSizedStructure(VariableSizedStructure s);
-    }
-    public void testVariableSizedStructureArgument() {
-        VariableSizeTest lib = (VariableSizeTest)Native.loadLibrary("testlib", VariableSizeTest.class);
-        String EXPECTED = getName();
-        VariableSizedStructure s = new VariableSizedStructure(EXPECTED);
-        assertEquals("Wrong string returned from variable sized struct",
-                     EXPECTED, lib.returnStringFromVariableSizedStructure(s));
-    }
-
     public void testNativeMappedWrite() {
     	class TestStructure extends Structure {
-    		public ByteByReference ref;
+            public ByteByReference ref;
     	}
     	TestStructure s = new TestStructure();
+        ByteByReference ref = s.ref = new ByteByReference();
+        s.write();
+        assertEquals("Value not properly written", ref.getPointer(), s.getPointer().getPointer(0));
+
     	s.ref = null;
     	s.write();
+        assertNull("Non-null value was written: " + s.getPointer().getPointer(0), s.getPointer().getPointer(0));
+    }
+
+    public void testNativeMappedRead() {
+    	class TestStructure extends Structure {
+            public ByteByReference ref;
+    	}
+    	TestStructure s = new TestStructure();
+        s.read();
+        assertNull("Should read null for initial field value", s.ref);
+
+        ByteByReference ref = new ByteByReference();
+        s.getPointer().setPointer(0, ref.getPointer());
+        s.read();
+        assertEquals("Field incorrectly read", ref, s.ref);
+
+        s.getPointer().setPointer(0, null);
+        s.read();
+        assertNull("Null field incorrectly read", s.ref);
     }
 
     public static class ROStructure extends Structure {
         public final int field;
         {
+            // Initialize in ctor to avoid compiler replacing
+            // field references with a constant everywhere
             field = 0;
-            getPointer().setInt(0, 42);
-            read();
         }
     }
+    private ROStructure avoidConstantFieldOptimization(ROStructure s) {
+        return s;
+    }
     public void testReadOnlyField() {
         ROStructure s = new ROStructure();
-        assertEquals("Field value should be writable from native", 42, s.field);
+        s.getPointer().setInt(0, 42);
+        s.read();
+        s = avoidConstantFieldOptimization(s);
+        assertEquals("Field value should be set from native", 42, s.field);
+
+        s.getPointer().setInt(0, 0);
+        s.read();
+        s = avoidConstantFieldOptimization(s);
+        assertEquals("Field value not synched after native change", 0, s.field);
+
+        s.getPointer().setInt(0, 42);
+        // current Java field value of zero should not be written to native mem
+        s.write();
+        assertEquals("Field should not be written", 42, s.getPointer().getInt(0));
+
+        s.read();
+        s = avoidConstantFieldOptimization(s);
+        assertEquals("Field value not synched after native change (2)", 42, s.field);
+
     }
     public void testNativeMappedArrayField() {
         final int SIZE = 24;
@@ -894,36 +980,66 @@ public class StructureTest extends TestCase {
         assertEquals("Wrong field value (2)", 0, ts.uninitialized.longValue());
     }
 
-    public void testStructureFieldOrder() {
-        Structure.REQUIRES_FIELD_ORDER = true;
-        try {
-            class TestStructure extends Structure {
-                public int one = 1;
-                public int three = 3;
-                public int two = 2;
-                {
-                    setFieldOrder(new String[] { "one", "two", "three" });
-                }
-            }
-            class DerivedTestStructure extends TestStructure {
-                public int four = 4;
-                {
-                    setFieldOrder(new String[] { "four" });
-                }
-            }
+    public void testInheritedStructureFieldOrder() {
+        class TestStructure extends Structure {
+            public int first = 1;
+        }
+        class TestStructureSub extends TestStructure {
+            public int second = 2;
+        }
+        TestStructureSub s = new TestStructureSub();
+        assertEquals("Wrong size", 8, s.size());
+        s.write();
+        assertEquals("Wrong first field: " + s,
+                     s.first, s.getPointer().getInt(0));
+        assertEquals("Wrong second field: " + s,
+                     s.second, s.getPointer().getInt(4));
+    }
 
-            DerivedTestStructure s = new DerivedTestStructure();
-            DerivedTestStructure s2 = new DerivedTestStructure();
-            s.write();
-            s2.write();
-            assertEquals("Wrong first field", 1, s.getPointer().getInt(0));
-            assertEquals("Wrong second field", 2, s.getPointer().getInt(4));
-            assertEquals("Wrong third field", 3, s.getPointer().getInt(8));
-            assertEquals("Wrong derived field", 4, s.getPointer().getInt(12));
+    public void testExplicitStructureFieldOrder() {
+        final String[] ORDER = new String[] { "one", "two", "three" };
+        final String[] ORDER2 = new String[] { "one", "two", "three", "four" };
+        class TestStructure extends Structure {
+            public int one = 1;
+            public int three = 3;
+            public int two = 2;
+            {
+                setFieldOrder(ORDER);
+            }
+            public List getFieldOrder() {
+                return super.getFieldOrder();
+            }
         }
-        finally {
-            Structure.REQUIRES_FIELD_ORDER = false;
+        class DerivedTestStructure extends TestStructure {
+            public int four = 4;
+            {
+                setFieldOrder(new String[] { "four" });
+            }
         }
+        
+        TestStructure s = new TestStructure();
+        assertEquals("Wrong field order",
+                     Arrays.asList(ORDER), s.getFieldOrder());
+        s.write();
+        assertEquals("Wrong first field: " + s,
+                     s.one, s.getPointer().getInt(0));
+        assertEquals("Wrong second field: " + s,
+                     s.two, s.getPointer().getInt(4));
+        assertEquals("Wrong third field: " + s,
+                     s.three, s.getPointer().getInt(8));
+
+        DerivedTestStructure s2 = new DerivedTestStructure();
+        assertEquals("Wrong field order",
+                     Arrays.asList(ORDER2), s2.getFieldOrder());
+        s2.write();
+        assertEquals("Wrong first field: " + s2,
+                     s2.one, s2.getPointer().getInt(0));
+        assertEquals("Wrong second field: " + s2,
+                     s2.two, s2.getPointer().getInt(4));
+        assertEquals("Wrong third field: " + s2,
+                     s2.three, s2.getPointer().getInt(8));
+        assertEquals("Wrong derived field: " + s2,
+                     s2.four, s2.getPointer().getInt(12));
     }
 
     public void testCustomTypeMapper() {
@@ -960,25 +1076,175 @@ public class StructureTest extends TestCase {
         assertNotNull("Field should not be null after read", s.field);
     }
 
-    public interface AutoSynchTest extends Library {
-        public static class TestStructure extends Structure {
+    public void testStructureEquals() {
+        class OtherStructure extends Structure {
+            public int first;
+            public int[] second = new int[4];
+            public Pointer[] third = new Pointer[4];
+        }
+        class TestStructure extends Structure {
+            public int first;
+            public int[] second = new int[4];
+            public Pointer[] third = new Pointer[4];
+        }
+        OtherStructure s0 = new OtherStructure();
+        TestStructure s1 = new TestStructure();
+        TestStructure s2 = new TestStructure();
+        TestStructure s3 = new TestStructure();
+        int VALUE = 99;
+        s1.first = s2.first = s3.first = VALUE;
+
+        assertFalse("Structures of different classes with same fields are not equal", s1.equals(s0));
+        assertFalse("Structures of different classes with same fields are not equal (reflexive)", s0.equals(s1));
+
+        assertFalse("Compare to null failed", s1.equals(null));
+        assertTrue("Equals is not reflexive", s1.equals(s1));
+        assertTrue("Equals failed on identical structures", s1.equals(s2));
+        assertTrue("Equals is not symmetric", s2.equals(s1));
+        assertTrue("Equals is not transitive", s1.equals(s2) && s2.equals(s3) && s1.equals(s3));
+
+        
+    }
+
+    public void testStructureEqualsByValueByReference() {
+        class TestStructure extends Structure {
+            public int first;
+            public int[] second = new int[4];
+            public Pointer[] third = new Pointer[4];
+        }
+        class ByReference extends TestStructure implements Structure.ByReference { }
+        class ByValue extends TestStructure implements Structure.ByValue { }
+        TestStructure s1 = new TestStructure();
+        TestStructure s2 = new ByReference();
+        TestStructure s3 = new ByValue();
+        int VALUE = 99;
+        s1.first = s2.first = s3.first = VALUE;
+
+        assertTrue("Equals failed on identical ByReference", s1.equals(s2));
+        assertTrue("Equals is not symmetric (ByReference)", s2.equals(s1));
+        assertTrue("Equals failed on identical ByValue", s1.equals(s3));
+        assertTrue("Equals is not symmetric (ByValue)", s3.equals(s1));
+        assertTrue("Equals is not transitive (ByReference/ByValue)", s1.equals(s2) && s2.equals(s3) && s1.equals(s3));
+
+        
+    }
+
+    public void testStructureHashCodeMatchesEqualsTrue() {
+        class TestStructure extends Structure {
+            public int first;
+        }
+        TestStructure s1 = new TestStructure();
+        TestStructure s2 = new TestStructure();
+        s1.first = s2.first = 0x12345678;
+        assertEquals("hashCode should match when structures equal",
+                     s1.hashCode(), s2.hashCode());
+    }
+
+    public void testStructureEqualsIgnoresPadding() {
+        class TestStructure extends Structure {
+            public byte first;
+            public int second;
+        }
+        TestStructure s1 = new TestStructure();
+        TestStructure s2 = new TestStructure();
+
+        // Make padding bits non-zero
+        s2.getPointer().setInt(0, -1);
+        s2.write();
+        assertTrue("Structure equals should ignore padding", s1.equals(s2));
+    }
+
+    public void testRecursiveWrite() {
+        class TestStructureByRef extends Structure implements Structure.ByReference{
+            public TestStructureByRef(Pointer p) { super(p); }
+            public TestStructureByRef() { }
+            public int unique;
+            public TestStructureByRef s;
+        }
+        TestStructureByRef s = new TestStructureByRef();
+        s.s = new TestStructureByRef();
+        s.unique = 1;
+        s.s.s = s;
+        s.s.unique = 2;
+
+        s.write();
+        assertEquals("Structure field contents not written",
+                     1, s.getPointer().getInt(0));
+        assertEquals("ByReference structure field contents not written",
+                     2, s.s.getPointer().getInt(0));
+
+        s.s.unique = 0;
+        Structure value = s.s;
+        s.read();
+        assertEquals("ByReference structure field not preserved", value, s.s);
+
+        assertEquals("ByReference structure field contents not read",
+                     2, s.s.unique);
+
+        assertTrue("Temporary storage should be cleared",
+                   s.busy().isEmpty());
+    }
+
+    public static class CyclicTestStructure extends Structure {
+        public static class ByReference extends CyclicTestStructure implements Structure.ByReference {}
+        public CyclicTestStructure(Pointer p) { super(p); }
+        public CyclicTestStructure() { }
+        public CyclicTestStructure.ByReference next;
+    }
+    public void testCyclicRead() {
+        CyclicTestStructure s = new CyclicTestStructure();
+        s.next = new CyclicTestStructure.ByReference();
+
+        Structure value = s.next;
+        s.next.next = s.next;
+        s.write();
+        s.read();
+        assertEquals("ByReference structure field not preserved", value, s.next);
+
+        value = s.next;
+        s.next.next = null;
+        s.read();
+        assertSame("ByReference structure field should reuse existing value",
+                   value, s.next);
+        assertSame("Nested ByReference structure field should reuse existing value",
+                   value, s.next.next);
+    }
+
+    public void testAvoidMemoryAllocationInPointerCTOR() {
+        class TestStructure extends Structure {
+            public int field;
+            public TestStructure(Pointer p) {
+                super(p);
+            }
+            protected Memory autoAllocate(int size) {
+                fail("Memory should not be auto-allocated");
+                return null;
+            }
+        }
+        Memory p = new Memory(4);
+        Structure s = new TestStructure(p);
+    }
+
+    public void testPointerCTORWithInitializedFields() {
+        class TestStructure extends Structure {
+            public byte[] field = new byte[256];
+            public TestStructure(Pointer p) {
+                super(p);
+            }
+        }
+        Memory p = new Memory(256);
+        Structure s = new TestStructure(p);
+        assertEquals("Wrong structure size", p.size(), s.size());
+    }
+
+    public void testEquals() {
+        class TestStructure extends Structure {
             public int field;
         }
-        Pointer testStructurePointerArgument(TestStructure s);
+        Structure s = new TestStructure();
+        assertTrue("Should match self", s.equals(s));
+        assertFalse("Not equal null", s.equals(null));
+        assertFalse("Not equal some other object", s.equals(new Object()));
     }
-    public void testDisableAutoSynch() {
-        AutoSynchTest lib = (AutoSynchTest)Native.loadLibrary("testlib", AutoSynchTest.class);
-        AutoSynchTest.TestStructure s = new AutoSynchTest.TestStructure();
-        final int VALUE = 42;
-        s.field = VALUE;
-        s.setAutoWrite(false);
-        lib.testStructurePointerArgument(s);
-        assertEquals("Auto write should be disabled", 0, s.field);
 
-        final int EXPECTED = s.field;
-        s.getPointer().setInt(0, VALUE);
-        s.setAutoRead(false);
-        lib.testStructurePointerArgument(s);
-        assertEquals("Auto read should be disabled", EXPECTED, s.field);
-    }
-}
\ No newline at end of file
+}
