@@ -21,15 +21,14 @@ import java.util.HashMap;
 import java.util.Map;
 import junit.framework.TestCase;
 
+@SuppressWarnings("unused")
 public class TypeMapperTest extends TestCase {
     public static interface TestLibrary extends Library {
         int returnInt32Argument(boolean b);
         int returnInt32Argument(String s);
         int returnInt32Argument(Number n);
     }
-    public TypeMapperTest(String testName) {
-        super(testName);
-    }
+
     public void testBooleanToIntArgumentConversion() {
         final int MAGIC = 0xABEDCF23;
         Map options = new HashMap();
@@ -159,7 +158,7 @@ public class TypeMapperTest extends TestCase {
         mapper.addTypeConverter(Boolean.class, converter);
         Map options = new HashMap();
         options.put(Library.OPTION_TYPE_MAPPER, mapper);
-        StructureTestLibrary lib = (StructureTestLibrary)
+		StructureTestLibrary lib = (StructureTestLibrary)
             Native.loadLibrary("testlib", StructureTestLibrary.class, options);
         StructureTestLibrary.TestStructure s = new StructureTestLibrary.TestStructure(mapper);
         assertEquals("Wrong native size", 4, s.size());
